# $BEGIN_AUDIOMATH_LICENSE$
# 
# This file is part of the audiomath project, a Python package for
# recording, manipulating and playing sound files.
# 
# Copyright (c) 2017-2019 Jeremy Hill
# 
# audiomath is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/ .
# 
# $END_AUDIOMATH_LICENSE$

"""
The GenericInterface submodule provides `GenericPlayer`. and
`GenericRecorder` classes. Specific interface implementations
can subclass these (for example, `PortAudioInterface`).
"""
__all__ = [
	'Fader',
]

import sys
import time
import inspect
import weakref

from . import Dependencies;         from .Dependencies import numpy

from . import Base; from .Base import Sound

class GenericPlayer( object ):

	__t0_global = [ None ]
	
	def __init__( self, sound=None ):
		if not isinstance( sound, Sound ): sound = Sound( sound )
		self.sound = sound
		self._loop = False
		self._playing = False
		self._speed = 1.0
		self._volume = 1.0
		self._levels = []
		self._pan = 0.0
		self._norm = 'inf'
		self.__t0_synch = None
		self.__t0_unsynch = None
		self._nextSample = 0
		self._blank = None
		self._blankParams = None
		self._channelSub = None
		self._channelSubParams = None
		self._sampleSub = None
		self._sampleSubParams = None
		self._sampleRateEqualizationFactor = 1.0
		self.__synchronizeDynamics = True
		self.__panLevels = [ 1.0, 1.0 ]

	def _OutputCallback( self, t, outputFrameCount, numberOfChannels, sampleFormat ):
		
		if self.__t0_global[ 0 ] is None: self.__t0_global[ 0 ] = t # all players will share this t0
		if self.__t0_synch is None: self.__t0_synch = t
		if self.__t0_unsynch is None: self.__t0_unsynch = t
		
		runDynamics = True
		if self.__synchronizeDynamics:
			runDynamics = self._playing
			if self.__synchronizeDynamics == 'playing' and not self._playing:
				self.__synchronizeDynamics = ( 'paused', t )
			elif isinstance( self.__synchronizeDynamics, tuple ) and self._playing:
				self.__t0_synch += t - self.__synchronizeDynamics[ 1 ]
				self.__synchronizeDynamics = 'playing'
			t -= self.__t0_synch
		else:
			t -= self.__t0_unsynch
		if runDynamics: self._RunDynamics( t )

		if not self._playing or not self._speed:
			return None

		blankParams = ( outputFrameCount, numberOfChannels, sampleFormat )
		if self._blankParams != blankParams:
			self._blankParams = blankParams
			self._blank = numpy.zeros( [ outputFrameCount, numberOfChannels ], dtype=sampleFormat )
		frame = self._blank

		speed = self._speed * self._sampleRateEqualizationFactor
		sampleSubParams = ( outputFrameCount, speed )
		if speed == 1.0:
			if self._sampleSubParams != sampleSubParams:
				self._sampleSubParams = sampleSubParams
				self._sampleSub = numpy.arange( outputFrameCount ) # NB: not a slice (see below)
			sampleSub = self._sampleSub + int( self._nextSample )
			nextSample = sampleSub[ -1 ] + 1
		else:
			if self._sampleSubParams != sampleSubParams:
				self._sampleSubParams = sampleSubParams
				self._sampleSub = numpy.linspace( 0, outputFrameCount * speed, outputFrameCount, endpoint=False )
			sampleSub = ( self._sampleSub + self._nextSample ).astype( int )
			nextSample = self._nextSample + outputFrameCount * speed
		
		snd = self.sound
		nSamplesAvailable = snd.NumberOfSamples()		
		if self._loop:
			sampleSub %= nSamplesAvailable
		elif sampleSub[ -1 ] >= nSamplesAvailable:
			sampleSub = sampleSub[ sampleSub < nSamplesAvailable ]
			frame.flat = 0 # silence the frame (although the beginning may be overwritten below)
			nextSample = 0
			self.playing = False
		source = snd.y[ sampleSub, : ] # this creates a copy, because sampleSub is a sequence not a slice
		
		nChannelsAvailable = source.shape[ 1 ]
		channelSubParams = ( numberOfChannels, nChannelsAvailable )
		if self._channelSubParams != channelSubParams:
			self._channelSubParams = channelSubParams
			if   numberOfChannels == nChannelsAvailable: self._channelSub = slice( None )
			elif numberOfChannels <  nChannelsAvailable: self._channelSub = slice( None, numberOfChannels )
			else: self._channelSub = ( list( range( nChannelsAvailable ) ) * numberOfChannels )[ :numberOfChannels ]
		channelSub = self._channelSub
		if channelSub: source = source[ :, channelSub ]
		
		left, right = self.__panLevels
		left *= self._volume
		right *= self._volume
		if left  != 1.0: source[ :, 0::2 ] *= left
		if right != 1.0: source[ :, 1::2 ] *= right
		levels = self._levels
		if levels:
			nch = source.shape[ 1 ]
			if len( levels ) != nch: levels[ : ] = ( levels * nch )[ :nch ]
			source *= [ levels ]
			
		# TODO: should apply the appropriate scaling factor when changing formats between float and int.
		# The logic is already implemented in `snd.dat2str( data=source, dtype=sampleFormat )` 
	    # but it really needs to happen outside this callback,  in `_wrap_portaudio._objectinterface` after data
	    # from multiple players have been mixed. Also its impacts on accuracy and efficiency are untested as yet.
	    
		frame[ :source.shape[ 0 ], :source.shape[ 1 ] ] = source # TODO: optimize packing of frame when created?
		self._nextSample = nextSample
		return frame
		
	def ResetDynamics( self, global_t0=None ):
		"""
		DOC-TODO
		"""
		if self.__synchronizeDynamics:
			self.__t0_synch = None
		else:
			self.__t0_unsynch = self.__t0_global[ 0 ] if global_t0 else None
	
	@classmethod
	def ResetGlobalClock( cls ):
		cls.__t0_global[ 0 ] = None

	@property
	def synchronizeDynamics( self ):
		"""
		Many properties, such as `.volume`, `.speed`, `.pan`,
		`.levels` and even `.playing`, can be dynamic, i.e.
		they can be set to functions of time `t` in seconds.
		
		If `p.synchronizeDynamics` is true, then the dynamic
		clock pauses whenever playback pauses and resumes only
		when `.playing` becomes true again. Hence, if `.playing`
		itself is dynamic and becomes false, it will not be
		able to become true again on its own). So this example
		will only play once::
		
			p.Play(0, loop=True, synchronizeDynamics=True, playing=lambda t:(t/p.sound.duration)%2<1 )
		
		If `p.synchronizeDynamics` is false, then the dynamic
		clock continues even when playback is paused. Hence,
		if `.playing` is dynamic, the `Player` can stop *and*
		resume on a schedule. So this example will play
		repeatedly, with pauses::
		
			p.Play(0, loop=True, synchronizeDynamics=False, playing=lambda t:(t/p.sound.duration)%2<1 )
		
		A separate but related issue is that of the "time zero"
		from which dynamic time `t` is measured.  By default,
		time zero for a given `Player` instance `p` is the time
		at which `p` was created. The clock can be reset by
		saying `p.ResetDynamics()`.  You can also synchronize
		`p` with a global clock, available to all `Player`
		instances, by saying `p.ResetDynamics('global')`		
		"""
		return True if self.__synchronizeDynamics else False
	@synchronizeDynamics.setter
	def synchronizeDynamics( self, value ):
		if value:
			if self.__synchronizeDynamics: return
			self.__synchronizeDynamics = 'playing'
			self.__t0_synch = None
		else:
			if not self.__synchronizeDynamics: return
			self.__synchronizeDynamics = False
			self.__t0_unsynch = None
			
		if not value and self.__synchronizeDynamics: return
	
	@property
	def playing( self ):
		"""
		DOC-TODO
		"""
		return self._playing
	@playing.setter
	def playing( self, value ):
		if self.SetDynamic( 'playing', value if callable( value ) else None ): return
		self._playing = value
	
	@property
	def loop( self ):
		"""
		DOC-TODO
		"""
		return self._loop
	@loop.setter
	def loop( self, value ):
		if self.SetDynamic( 'loop', value if callable( value ) else None ): return
		self._loop = value
	
	@property
	def speed( self ):
		"""
		DOC-TODO
		"""
		return self._speed
	@speed.setter
	def speed( self, value ):
		if self.SetDynamic( 'speed', value if callable( value ) else None ): return
		self._speed = float( value )
		
	@property
	def volume( self ):
		"""
		DOC-TODO
		"""
		return self._volume
	@volume.setter
	def volume( self, value ):
		if self.SetDynamic( 'volume', value if callable( value ) else None ): return
		self._volume = float( value )
	vol = volume
	
	@property
	def levels( self ):
		"""
		DOC-TODO
		"""
		return self._levels
	@levels.setter
	def levels( self, value ):
		if self.SetDynamic( 'levels', value if callable( value ) else None ): return
		try: len( value )
		except: value = [ float( value ) ]
		else: value = [ float( each ) for each in value ]
		self._levels[ : ] = value

	@property
	def pan( self ):
		"""
		DOC-TODO
		"""
		return self._pan
	@pan.setter
	def pan( self, value ):
		if self.SetDynamic( 'pan', value if callable( value ) else None ): return
		try: len( value )
		except: pass
		else: self.__panLevels[ : ] = float( value[ 0 ] ), float( value[ 1 ] ); return
		self._pan = float( value )
		self._UpdateLevels( pan=value )

	@property
	def norm( self ):
		"""
		DOC-TODO
		"""
		return self._norm
	@norm.setter
	def norm( self, value ):
		if self.SetDynamic( 'norm', value if callable( value ) else None ): return
		self._norm = value
		self._UpdateLevels( norm=value )
		
	def _UpdateLevels( self, pan=None, norm=None ):
		if pan  is None: pan  = self._pan
		if norm is None: norm = self._norm
		if norm is None or pan is None: return
		v = [ 0.5 + 0.5 * min( 1.0, max( -1.0, vi ) ) for vi in [ -pan, pan ] ]
		if 'inf' in str( norm ).lower():  denom = max( v, key=abs )
		else: denom = sum( vi ** norm for vi in v ) ** ( 1.0 / norm if norm else 1.0 )
		if denom: v = [ vi / float( denom ) for vi in v ]
		self.__panLevels[ : ] = v
	
	def GetDynamic( self, name ):
		"""
		DOC-TODO
		"""
		dynamics = getattr( self, '_dynamics', None )
		return dynamics.get( name, None ) if dynamics else None
		
	def SetDynamic( self, name, func, order=1 ):
		"""
		DOC-TODO
		"""
		dynamics = getattr( self, '_dynamics', None )
		if func is None:
			if dynamics: dynamics.pop( name, None )
		else:
			if dynamics is None: dynamics = self._dynamics = {}
			try: getargspec = inspect.getfullargspec
			except: getargspec = inspect.getargspec
			try: spec = getargspec( func )
			except: spec = getargspec( func.__call__ )
			nArgs = len( spec.args )
			if hasattr( func, '__self__' ): nArgs -= 1
			elif not inspect.isfunction( func ) and not inspect.ismethod( func ): nArgs -= 1  # callable custom instance: won't have a __self__ itself but its __call__ method will
			if nArgs == 0 and spec.varargs: nArgs = 1
			dynamics[ name ] = ( order, name, func, nArgs )
			return func
	
	def _RunDynamics( self, t ):
		dynamics = getattr( self, '_dynamics', None )
		if dynamics is None: dynamics = self._dynamics = {}
		updates = {}
		for entry in sorted( dynamics.values() ):
			order, name, func, nArgs = entry
			try:
				if nArgs == 0: value = func()
				elif nArgs == 1: value = func( t )
				elif nArgs == 2: value = func( self, t )
				else: raise TypeError( 'dynamic property values should take 0, 1 or 2 arguments' )
			except StopIteration as exc:
				value = None
				for arg in exc.args:
					if isinstance( arg, dict ):
						if value is None: value = arg.pop( '_', None )
						updates.update( arg )
					else: value = arg
				if value is None: dynamics.pop( name )
				else: setattr( self, name, value )
			except:
				einfo = sys.exc_info()
				sys.stderr.write( 'Exception while evaluating dynamic .%s property of %s:\n' % ( name, self ) )
				getattr( self, '_excepthook', sys.excepthook )( *einfo )
				dynamics.pop( name )
			else:
				setattr( self, name, value )
				dynamics[ name ] = entry
		if updates:
			for name, value in updates.items():
				setattr( self, name, value )

	def Set( self, **kwargs ):
		"""DOC-TODO"""
		for k, v in kwargs.items():
			if not hasattr( self, k ): raise AttributeError( '%s instance has no attribute %r' % ( self.__class__.__name__, k ) )
			setattr( self, k, v )
		return self
	
	def Resample( self, newFs ):
		"""
		Replace `self.sound` with a copy resampled at the
		new sampling rate `newFs`.  Also adjust the Player's
		internal speed compensation factor, to ensure things
		sound the same on the existing stream. The most
		common usage will be to match the sound sample rate
		to the sample rate of the hardware (so the Player's
		compensation factor becomes 1.0) to ensure more
		efficient playback.
		"""
		if hasattr( newFs, 'fs' ): newFs = newFs.fs
		oldFs = self.sound.fs
		self.sound = self.sound.Copy().Resample( newFs )
		self._sampleRateEqualizationFactor *= newFs / oldFs 
		
	def Play( self, position=None, wait=False, **kwargs ):
		"""
		Start playing. Optionally use `position` to specify
		the exact `.head` position from which to start playing
		(in seconds from the beginning). Otherwise playack
		starts from wherever it last ended up (or at the
		beginning, if the sound previously ran to the end).
		
		If you call `p.Play()` without an explicit `position`
		on an instance `p` that is *already* playing, nothing
		changes---playback simply continues. In this sense it
		is equivalent to manipulating the `.playing` property
		by saying `p.playing = True`.
		 
		If `wait` is true, immediately call `.Wait()` so that
		playback is synchronous.
		
		Optionally, use further keyword arguments to set
		additional properties (e.g. `volume`, `pan`, `levels`,
		`speed`, `loop`, etc) at the same time.
		"""
		if position is not None: self.Seek( position )
		if kwargs: self.Set( **kwargs )
		self._playing = True
		if wait: self.Wait()
		#return self
		
	def Pause( self, position=None, **kwargs ):
		"""
		Stop playing. Optionally change the `.head` position
		to the specified absolute `position` in seconds, ready
		for subsequent playback. Optionally, use keyword
		arguments to set additional properties (e.g. `volume`,
		`pan`, `levels`, `speed`, `loop`, etc) at the same time.
		"""
		if position is not None: self.Seek( position )
		if kwargs: self.Set( **kwargs )
		self._playing = False
		#return self
	Stop = Pause
	
	def Wait( self ):
		"""
		Sleep until `self` stops playing, or until the user
		presses ctrl-C to interrupt.
		"""
		try:
			while self.playing: time.sleep( 0.001 )
		except KeyboardInterrupt:
			self.playing = False

	def Seek( self, position, relative=False ):
		"""
		Move the playback `.head` to the specified `position`
		in seconds. Negative values count back from the end of
		`self.sound`. If `relative` is true, `position` is
		interpreted as an offset relative to the current head
		position. The following are equivalent::
		
			p.Seek( 5.0, relative=False )  # more efficient
			p.head = 5.0
		
		And the following are also equivalent:
		
			p.Seek( -2.0,  relative=True ) # more efficient
			p.head -= 2.0
		"""
		samples = int( round( self.sound.fs * position ) )
		if relative: self._nextSample = max( 0, self._nextSample + samples )
		else: self._nextSample = ( self.sound.NumberOfSamples() + samples ) if samples < 0 else samples
		return self
	
	@property
	def head( self ):
		"""
		Current playback position, in seconds relative to the
		beginning.
		"""
		return self._nextSample / self.sound.fs
	@head.setter
	def head( self, value ):
		self.Seek( value, relative=False )
		

class GenericRecorder( object ):

	def __init__( self, seconds, fs=44100, numberOfChannels=1, sampleFormat='float32', start=True ):
		if isinstance( seconds, Sound ): self.sound = seconds
		else: self.sound = Sound( seconds, fs=fs, nchan=numberOfChannels )
		self._numberOfChannels = self.sound.NumberOfChannels()
		self._nextSample = 0
		self.recording = start
		self.__array_interface__ = dict( version=3, typestr=sampleFormat )
		
	def _InputCallback( self, t, inputData, frameCount ):
		if self.recording:
			self.__array_interface__[ 'data' ] = ( inputData.value, True )
			self.__array_interface__[ 'shape' ] =( frameCount, self._numberOfChannels )
			x = numpy.array( self )
			dst = self.sound.y[ self._nextSample:self._nextSample + frameCount, : ]
			dst.flat = x[ :dst.shape[ 0 ], : ].flat
			self._nextSample += dst.shape[ 0 ]
			if self._nextSample >= self.sound.y.shape[ 0 ]:
				self.recording = False
				self._nextSample = 0
		
	def Record( self, position=None, wait=True ):
		"""
		DOC-TODO
		"""
		if position is not None: self.Seek( position )
		self.recording = True
		if wait: self.Wait()
		
	def Pause( self, position=None ):
		"""
		DOC-TODO
		"""
		if position is not None: self.Seek( position )
		self.recording = False
	Stop = Pause
		
	def Wait( self ):
		"""
		DOC-TODO
		"""
		try: self.WaitFor( self )  # ownself wait for ownself lah!
		except KeyboardInterrupt: self.recording = False
	
	def WaitFor( self, condition, finalRecordingStatus=None ):
		"""
		Sleep until `condition` is fulfilled. The `condition`
		argument may be:
				
			a `Recorder` instance (most likely `self`):
				in which case the method waits until the
				recorder is no longer recording;
				
			a `Player` instance:
				in which case the method waits until the
				player is no longer playing;
				
			a string:
				in which case the method will use the
				string as a prompt and wait for the user
				to enter any string (even the empty string)
				on the console;
				
			any callable:
				in which case the method will repeatedly
				call `condition()` and wait for it to
				return anything other than `None`.
		
		If `finalRecordingStatus` is not `None`, then
		`self.recording` is set to its value when the wait
		is over.  So for example, you can prompt a user
		to end an ongoing recording as follows::
		
			r.WaitFor('Press return to stop recording: ', finalRecordingStatus=False)
		"""
		if isinstance( condition, GenericRecorder ):
			obj = condition
			condition = lambda: None if obj.recording else '%r finished recording' % obj
		if isinstance( condition, GenericPlayer ):
			obj = condition
			condition = lambda: None if obj.playing else '%r finished playing' % obj
		if isinstance( condition, str ):
			prompt = condition
			try: from builtins import raw_input as ask # Python 2 (for some reason just referencing `raw_input` directly gave a NameError on Python 2.7.15 on linux...)
			except ImportError: ask = input # Python 3
			def condition():
				try: return ask( prompt )
				except EOFError: print( '' ); return ''
		while True:
			result = condition()
			if result is not None: break
			time.sleep( 0.001 )
		if finalRecordingStatus is not None:
			self.recording = finalRecordingStatus
		return result
		
	def Seek( self, position, relative=False ):
		"""
		Move the recording `.head` to the specified `position`
		in seconds. Negative values count back from the end of
		the available recording space in `self.sound`. If
		`relative` is true, `position` is interpreted as an
		offset relative to the current head position. The
		following are equivalent::
		
			r.Seek( 5.0, relative=False )  # more efficient
			r.head = 5.0
		
		And the following are also equivalent:
		
			r.Seek( -2.0,  relative=True ) # more efficient
			r.head -= 2.0
		"""
		samples = int( round( self.sound.fs * position ) )
		if relative: self._nextSample = max( 0, self._nextSample + samples )
		else: self._nextSample = self.sound.NumberOfSamples() + samples if samples < 0 else samples
		return self
	
	@property
	def head( self ):
		"""
		Current recording position, in seconds relative to the
		beginning.
		"""
		return self._nextSample / self.sound.fs
	@head.setter
	def head( self, value ):
		self.Seek( value, relative=False )
		
	def Cut( self, position=None ):
		"""
		Stop recording, shorten `self.sound` by cutting it off at
		the current `.head` position (or the explicitly specified
		`position` in seconds), and return the `Sound` instance.
		"""
		if position is None: position = self.head
		if position: self.sound = self.sound[ :float( position ) ] # there was a reason for not cutting when position is 0
		return self.sound

	@classmethod
	def MakeRecording( cls, space=60, prompt='Recording for {duration:g} seconds - press return to finish early: ', verbose=None, cut=True, **kwargs ):
		"""
		Record and return a `Sound`.
		
		Args:
			space (Sound, float):
				specifies either a pre-existing `Sound` instance
				into which to record,  or a number of seconds. In
				the latter case a `Sound` instance is created, pre-
				allocated with the specified amount of space in
				seconds.
			
			prompt (str):
				If supplied, this provides console interaction:
				the prompt is printed, and you can press return
				to end the recording early. If empty or `None`,
				you can still stop the recording early with ctrl-C.
				
			verbose (bool):
				Passed through to the constructor of the `Recorder`
				object that is being used.
			
			cut (bool):
				If true, return a `.Cut()` version of the `Sound`
				when recording stops---i.e. ensure that the
				duration of the `Sound` does not exceed the
				duration of the recording.  If false, the entire
				pre-allocated `space` is returned, with the
				recording at the beginning.
				
			**kwargs:
				Additional keyword arguments are passed through to
				the constructor of the `Recorder` object (for example,
				to specify the `device` that should be used).
		"""
		r = cls( seconds=space, start=True, verbose=verbose, **kwargs )
		if prompt:
			if isinstance( prompt, str ): prompt = prompt.format( duration=r.sound.Duration() )
			r.WaitFor( prompt, finalRecordingStatus=False )
		else:
			if r.verbose: print( 'Recording for {duration:g} seconds - press ctrl + C to finish early'.format( duration=r.sound.Duration() ) )
			r.Wait()
		return r.Cut() if cut else r.sound

class Fader( object ):
	def __init__( self, duration=1.0, start=1, end=0, transform=None, pauseWhenDone='auto' ):
		"""DOC-TODO"""
		self.t0 = None
		self.start = start
		self.end = end
		self.duration = duration
		self.transform = transform
		if pauseWhenDone == 'auto': pauseWhenDone = ( end == 0 )
		self.pauseWhenDone = pauseWhenDone
	def __call__( self, t ):
	 	if self.t0 is None: self.t0 = t
	 	t -= self.t0
	 	if t <= self.duration: value = self.start + ( self.end - self.start ) * ( t / self.duration )
	 	elif self.pauseWhenDone: raise StopIteration( dict( playing=False ) )
	 	else: value = self.start * 0 + self.end 
	 	if self.transform: value = self.transform( value )
	 	return value
